/* SMTP Server state machine - see RFC 821
 *  enhanced 4/88 Dave Trulli nn2z
 */

/****************************************************************************
*	$Id: smtpserv.c 1.17 2001/06/14 09:49:08 Simon Exp $
*	25 Jun 92    paul@wolf.demon.co.uk added automated mail bouncing		*
*	13 Jul 92	1.3		GT	Remove 1.2 changes.								*
*   27 Aug 92   1.6     mt@kram.demon.co.uk added smtp separator            *
*	03 Sep 92	1.7		GT	Fix missing BOUNCER conditional.				*
*	08 Dec 92	1.8		mt@kram.org: Occasional pwait() during copy			*
*						cms@home:	 Alias-File Mail Bouncing System		*
*	19 Mar 93	1.9		GT	Debugging mailit.								*
*	04 Apr 93	1.10	GT	Re-enable beep.									*
*	05 Apr 93	1.11	GT	Reinstate delay call.							*
*	08 May 93	1.12	GT	Fix warnings.									*
*						IAY	Improve behaviour of rest of system				*
*							while this module is copying files,				*
*							particularly while in SMTP MODE QUEUE.			*
*						IAY	Fix dot transparency as per RFC 821.			*
*						GBD	Fix alias expansion to allow multiple spaces.	*
*	08 Dec 93	1.13	GT	Fix bounce message envelope From.				*
*							Version and compilation date in SMTP banner.	*
*							Implement VRFY.									*
*							Fix wandering pointer in getmsgtxt ().			*
*							Close mail spool file before ACK.				*
*							Fix "." at column 256.							*
*   30 Sep 95   1.14    MSM Fix local policy bounce from 250 to 503         *
*	   May 97	1.15	GeK	Accept tabs in commands, update ptime stamp fmt	*
*	         	    	   	mail notification moved, mail rejection logged	*
*	         	    	   	mailkill faster(?) matching, gives up time and	*
*	         	    	   	logs rule matched, always accept "postmaster"	*
*	         	    	   	"smtp relay" option, fixed default alias expn	*
*	         	    	   	EXPN & VRFY fixed to match alias expn			*
*	   Nov 97	1.16	GeK	Removed DOS file commit code; beep on mail		*
*	         	    	   	arrival not buffered; moved "From " escaping on	*
*	         	    	   	final mailbox delivery to copy_data (rewritten)	*
*	         	    	   	added !keep rules for mailkill; validate_sender	*
*	         	    	   	takes pointer for rule; mailkill logging tidied	*
*	         	    	   	added %-hack code; removed main_exit on pwaits;	*
*	         	    	   	250 Sent response includes message-id; smtp		*
*	         	    	   	close log shows message count; added pause &	*
*	         	    	   	retry on locked mailbox; sequence.seq is locked	*
*	         	    	   	when updated; alias expansion substantially		*
*	         	    	   	rewritten; unknown user takes preference over	*
*	         	    	   	mailkill; greater use of dynamic buffers;		*
*	         	    	   	handle up to 5 char TZ abbrev.					*
*	14 Jun 01	1.17	SCT	Escape "Status:" header line in mailit, turn	*
*							BREAK_LONG_LINES off by default					*
****************************************************************************/

/*
** The following definition controls how often the copy_data
** function calls pwait().  Lower numbers mean better response
** to other things that may be going on at the expense of the
** speed of that particular function.
*/
#define COPY_DATA_WAIT 1

#define	MMDF_FIX				/*  GeK - fix MMDF's nasty habit of quoting domain literals  */
#define	BLANK_LINE_BETWEEN_MSGS	/*  GeK - PCElm doesn't require this, others might  */
#undef	BREAK_LONG_LINES		/*  GeK - adds \n every 255 chrs for ted.com  */
#undef	RECURSIVE_ALIAS			/*  GeK - define this for recursive aliases (untested)  */
#undef	WILDCARD_ALIAS			/*  GeK - define this for wildcard aliases (untested)  */
#undef	VERIFY_FROM_ADDRESS		/*  GeK - DNS lookups on incoming From envelopes  */
#define	COMMIT_CACHE			/*  GeK - flush SmartDrive cache before Ack'ing mail  */
#undef	SERVER_TRACE			/*  GeK - SMTP server tracing  */
#define	PERCENT_HACK_ROUTE_FIX	/*  GeK - % hack fix (for post.demon.co.uk)  */
#define	ESCAPE_STATUS_HDR		/*  SCT - Escape "Status:" header lines in new mail  */

#include <stdio.h>
#include <time.h>

#undef	DEBUG_LOCKS

#ifdef NOVELL

#include "nit.h"
#include <io.h>
#include <dos.h>
#include <dir.h>
#include <time.h>

extern char novell_server_name[128];
extern char novell_mail_ext[4];
extern unsigned short novell_start;

#endif	/* NOVELL */

extern char *smtp_separator;
extern int Smtpbeep;
extern int smtpverbose;		/*  GeK - used in a few places  */
#ifdef	PERCENT_HACK_ROUTE_FIX	/*  GeK  */
extern char *smtp_gateway;		/*  in smtpcli.c  */
#endif

#ifdef UNIX
#include <sys/types.h>
#endif
#if	defined(__STDC__) || defined(__TURBOC__)
#include <stdarg.h>
#endif
#include <ctype.h>
#include <setjmp.h>
#include "global.h"
#include "mbuf.h"
#include "cmdparse.h"
#include "socket.h"
#include "iface.h"
#include "proc.h"
#include "smtp.h"
#include "commands.h"
#include "dirutil.h"
#include "mailbox.h"
#include "bm.h"
#include "domain.h"
#include "files.h"
#include "ip.h"
#include "netuser.h"	/*  GeK - for inet_ntoa()  */
#include "pc.h"			/*  GeK - for smartdrive_commit_cache()  */

char *Days[7] =
	{
	 "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"
	}
    ;
char *Months[12] =
	{
	 "Jan", "Feb", "Mar", "Apr", "May", "Jun",
	 "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
	}
    ;

static void copy_data __ARGS ((FILE * from, FILE * to));
static struct list *expandalias __ARGS ((struct list ** head, char *user, failure *default_fail));
static int getmsgtxt __ARGS ((struct smtpsv * mp));
static struct smtpsv *mail_create __ARGS ((void));
static void mail_clean __ARGS ((struct smtpsv * mp));
static int mailit __ARGS ((FILE * data, char *from, struct list * tolist));
static int router_queue __ARGS ((FILE * data, char *from, struct list * to));
static void smtplog __ARGS ((char *fmt,...));
static void smtpserv __ARGS ((int s, void *unused, void *p));
static int mailuser __ARGS ((FILE * data, char *from, char *to));
static int validate_user __ARGS ((char *user));
static int validate_sender __ARGS ((char *sender, char **rule));
#ifdef	notdef  /*  GeK  */
static int remlist __ARGS ((struct list ** head, struct list * unwanted));
#endif	/* notdef */
static struct list *expand_local_alias (char *address);

#ifdef	COMMIT_CACHE
static int docommit __ARGS ((int argc, char *argv[], void *p));
#endif
static int doreceipt __ARGS ((int argc, char *argv[], void *p));
static int dorelay __ARGS ((int argc, char *argv[], void *p));
#ifdef	SERVER_TRACE
static int do_server_trace __ARGS ((int argc, char *argv[], void *p));
#endif

static struct cmds smtp_server_cmds[] =
{
#ifdef	COMMIT_CACHE
	{ "commit",		docommit,	0,	0,	NULL },
#endif
	{ "receipt",	doreceipt,		0,	0,	NULL },
	{ "relay",		dorelay,		0,	0,	NULL },
#ifdef	SERVER_TRACE
	{ "trace",		do_server_trace,	0,	0,	NULL },
#endif
	{ NULL }
};

/* Command table */
static char *commands[] =
	{
	 "helo",
#define	HELO_CMD	0
	 "noop",
#define	NOOP_CMD	1
	 "mail from:",
#define	MAIL_CMD	2
	 "quit",
#define	QUIT_CMD	3
	 "rcpt to:",
#define	RCPT_CMD	4
	 "help",
#define	HELP_CMD	5
	 "data",
#define	DATA_CMD	6
	 "rset",
#define	RSET_CMD	7
	 "expn",
#define EXPN_CMD	8
	 "vrfy",
#define	VRFY_CMD	9
	 "ehlo",
#define	EHLO_CMD	10
	 NULLCHAR
	}
    ;

/* Reply messages */
static char Help[] = "214-Commands:\n214-HELO NOOP MAIL QUIT RCPT HELP DATA RSET EXPN VRFY\n214 End\n";
static char Banner[] = "220 %s %s SMTP ready\n";
static char Closing[] = "221 Closing\n";
static char Ok[] = "250 <%s> OK\n";
static char Reset[] = "250 Reset state\n";
static char Sent[] = "250 Sent (id AA%ld)\n";
static char Ourname[] = "250 %s Pleased to meet you %s ([%s])\n";	/*  GeK - added addr, drop comma (draft reqs space after hostname)  */
static char Unwanted[] = "250 Local policy blocks mail from <%s>\n";
static char Enter[] = "354 Enter mail, end with .\n";
static char Ioerr[] = "452 Temp file write error\n";
static char Badcmd[] = "500 Command unrecognized\n";
static char Lowmem[] = "421 System overloaded, try again later\n";
static char Syntax[] = "501 Syntax error\n";
static char Needrcpt[] = "503 Need valid RCPT (recipient)\n";
#ifdef	notdef	/*  GeK  */
static char Needsender[] = "503 Local policy blocking mail from you\n";
#endif
static char Unknown[] = "550 <%s> address unknown\n";
static char Noalias[] = "550 No alias for <%s>\n";
static char UnknownRcpt[] = "550 Unknown recipient <%s>\n";
static char NoPermission[] = "550 Permission denied; local policy blocks <%s>\n";
static char NoRelay[] = "551 Address not local; please try <%s>\n";
static char WillForward[] = "251 Address not local; will forward to <%s>\n";

#ifdef	notdef	/*  GeK - now unused globals  */
static failure fail;			/*  GeK - this is now just local to expand_local_alias()  */
static int bad_sender;			/*  GeK - from NJL - should be one-per-session  */
char default_address[SLINELEN];
#endif	/* notdef */

#ifdef	COMMIT_CACHE
static int	commit_cache = FALSE;	/*  GeK - flush SmartDrive cache  */
#endif
static int	smtp_relay = TRUE;		/*  GeK - smtp relaying [on | off]  */
static int	smtp_receipt = TRUE;	/*  GeK - smtp (return-) receipt [on | off]  */
#ifdef	SERVER_TRACE
static int	smtp_trace = 0;			/*  GeK - server tracing  */
#endif

static int Ssmtp = -1;		 /* prototype socket for service */

/* Start up SMTP receiver service */
int
    smtp1 (argc, argv, p)
int argc;
char *argv[];
void *p;
	{
	struct sockaddr_in lsocket;
	int s;

	if (Ssmtp != -1)
		{
		return 0;
		}

	psignal (Curproc, 0);						 /* Don't keep the parser
												  * waiting */
	chname (Curproc, "SMTP listener");

	lsocket.sin_family = AF_INET;
	lsocket.sin_addr.s_addr = INADDR_ANY;
	if (argc < 2)
		lsocket.sin_port = IPPORT_SMTP;
	else
		lsocket.sin_port = atoi (argv[1]);

	Ssmtp = socket (AF_INET, SOCK_STREAM, 0);
	bind (Ssmtp, (char *) &lsocket, sizeof (lsocket));
	listen (Ssmtp, 1);
	for (;;)
		{
		if ((s = accept (Ssmtp, NULLCHAR, (int *) NULL)) == -1)
			break;								 /* Service is shutting down */

		if (availmem () < Memthresh)
			{
			usprintf (s, Lowmem);
			shutdown (s, 1);
			}
		else
			{
			/* Spawn a server */
			/*  GeK - reduced stack size from 2048  */
			newproc ("SMTP server", 1280, smtpserv, s, NULL, NULL, 0);
			}
		}

	return 0;
	}

/* Shutdown SMTP service (existing connections are allowed to finish) */
int
    smtp0 (argc, argv, p)
int argc;
char *argv[];
void *p;

	{
	close_s (Ssmtp);
	Ssmtp = -1;
	return 0;
	}

static void
     smtpserv (s, unused, p)
int s;
void *unused;
void *p;
	{
	struct smtpsv *mp;
	char **cmdp, buf[LINELEN], *arg, *cp, *cmd, *newaddr;
	struct list *ap, *alias_list;
	int cnt;
	char address_type;
	int	socksize = SOCKSIZE;		/*  GeK - used in HELO  */
	struct sockaddr_in sender_ip;	/*  GeK - used in HELO  */
	char *bad_sender = NULL;		/*  GeK - was global, set to matched mailkill rule  */
	unsigned int	mail_from_count = 0;	/*  GeK  */
	unsigned int	messages_received = 0;	/*  GeK  */
	char *original_envelope;		/*  GeK - copy of original envelope in RCPT TO  */

	sockmode (s, SOCK_ASCII);
	sockowner (s, Curproc);						 /* We own it now */
	log (s, "open SMTP");

#ifdef	notdef		/*  GeK - mail_create can't fail  */
	if ((mp = mail_create ()) == NULLSMTPSV)
		{
		tprintf (Nospace);
		log (s, "close SMTP - no space");
		close_s (s);
		return;
		}
#endif
	mp = mail_create ();

	mp->s = s;
	/*  GeK - record remote details here  */
	mp->sender_ip_addr = (getpeername (mp->s, (char *) &sender_ip, &socksize) < 0)
	                     ? 0
	                     : sender_ip.sin_addr.s_addr;

	(void) usprintf (s, Banner, Hostname, Version);

loop:
	if ((cnt = recvline (s, buf, sizeof (buf))) == -1)
		{
		/* He closed on us */
		goto quit;
		}

	if (cnt < 4)
		{
		/* Can't be a legal command */
		usprintf (mp->s, Badcmd);
		goto loop;
		}

	rip (buf);
	cmd = buf;

	/* Translate entire buffer to lower case */
	for (cp = cmd; *cp != '\0'; cp++)
		*cp = tolower (*cp);

	/* Find command in table; if not present, return syntax error */
	for (cmdp = commands; *cmdp != NULLCHAR; cmdp++)
		if (strncmp (*cmdp, cmd, strlen (*cmdp)) == 0)
			break;

	if (*cmdp == NULLCHAR)
		{
		(void) usprintf (mp->s, Badcmd);
		goto loop;
		}

	arg = &cmd[strlen (*cmdp)];
	/* Skip spaces after command */
	while (*arg == ' ' || *arg == '\t')		/*  GeK - include tab  */
		arg++;

	/* Execute specific command */
	switch (cmdp - commands)
		{
		case EHLO_CMD:
			mp->esmtp = TRUE;
			/*  Fall through  */
		case HELO_CMD:
			free (mp->system);
			mp->system = strdup (arg);
			/*  GeK - DRUMS draft makes clear that HELO should be same as RSET  */
			del_list (mp->to);
			mp->to = NULL;
			free (mp->envelope_to);		/*  GeK - reset first recipient  */
			mp->envelope_to = NULL;
			usprintf (mp->s, Ourname, Hostname, mp->system,
			          inet_ntoa (mp->sender_ip_addr));
			break;
		case NOOP_CMD:
			usprintf (mp->s, Ok, commands[NOOP_CMD]);
			break;
		case MAIL_CMD:
			free (bad_sender);
			bad_sender = NULL;
			if ((cp = getname (arg)) == NULLCHAR)
				{
				usputs (mp->s, Syntax);
				break;
				}

			mail_from_count++;		/*  GeK  */

#ifdef	MMDF_FIX
			/*  GeK - fix MMDF's habit of quoting RHS domain literals  */
			if ((newaddr = strchr (cp, '@')) != NULL
			    && (++newaddr)[0] == '\"'
			    && newaddr[1] == '[')
				{
				while ((*newaddr = *(newaddr+1)) != '\"' && *newaddr != '\0')
					newaddr++;
				*newaddr = '\0';
				}
#endif	/* MMDF_FIX */

			free (mp->from);
			mp->from = strdup (cp);

#ifdef	VERIFY_FROM_ADDRESS
			mp->from_verified = FALSE;
			if (*cp == '\0'
			    || ((newaddr = strchr (cp, '@')) != NULL
			        && strchr (++newaddr, '.') != NULL
			        && (stricmp (Hostname, newaddr) == 0 || resolve_mx (newaddr) != 0))
			    || newaddr == NULL)		/*  allow local addresses  */
				mp->from_verified = TRUE;
			else
				log (mp->s, "SMTP From failed: <%s>", cp);
#endif	/* VERIFY_FROM_ADDRESS */

			/*  GeK - reset these in case of improper command ordering  */
			del_list (mp->to);
			mp->to = NULLLIST;

			free (mp->envelope_to);		/*  GeK - reset first recipient  */
			mp->envelope_to = NULL;

			usprintf (mp->s, validate_sender (cp, &bad_sender) ? Ok : Unwanted, cp);

			break;
		case QUIT_CMD:
			usputs (mp->s, Closing);
			goto quit;
		case RCPT_CMD:
			/* Specify recipient */
			if ((cp = getname (arg)) == NULLCHAR)
				{
				usputs (mp->s, Syntax);
				break;
				}

			/*
			 *  GeK - strip any attempted source routeing here, there is no
			 *        requirement that we honour it
			 */
			if (*cp == '@' && (newaddr = strchr (cp, ':')) != NULL)
				cp = newaddr+1;

			/*  GeK - save original envelope before rewrite if first recipient  */
			original_envelope = strdup (cp);

			/* rewrite address if possible */

			if ((newaddr = rewrite_address (cp)) != NULLCHAR)
				{
				strcpy (buf, newaddr);
				cp = buf;
				free (newaddr);
				}

			/* check if address is ok */

			if ((address_type = validate_address (cp)) == BADADDR)
				{
				usprintf (mp->s, Unknown, cp);
				/*  GeK - added rejection logging  */
				smtplog ("unknown address rejected: To: %s From: %s", cp, mp->from);
				if (smtpverbose)
					tprintf ("SMTP: Rejecting unknown address %s from %s\n", cp, mp->from);

				free (original_envelope);
				break;	/*  case RCPT_CMD  */
				}

			/* if a local address check for an alias */

			if (address_type == LOCAL)
				{
				int   must_accept;

				must_accept = (stricmp ("Postmaster", cp) == 0);

				alias_list = expand_local_alias (cp);
				if (alias_list == NULL && !must_accept)
					{
					usprintf (mp->s, UnknownRcpt, cp);
					/*  rejection logging  */
					smtplog ("unknown user rejected: To: %s From: %s", cp, mp->from);

					if (smtpverbose)
						tprintf ("SMTP: Rejecting unknown user %s from %s\n", cp, mp->from);
					}
				else if (bad_sender != NULL && !must_accept)
					{
					usprintf (mp->s, NoPermission, mp->from);
					smtplog ("mailkill rejected: To: %s From: %s Rule: %s", cp, mp->from, bad_sender);
					if (smtpverbose)
						tprintf ("SMTP: Rejecting mail to %s from %s (rule %s)\n", cp, mp->from, bad_sender);
						
					/*  Delete alias_list  */
					del_list (alias_list);
					}
				else	/*  must_accept || (alias_list == NULL && bad_sender == NULL)  */
					{
					/*  If local alias failed and must accept, create list entry  */
					if (alias_list == NULL && must_accept)
						addlist (&mp->to, cp, address_type);
					else
						{
						/*  Transfer alias_list to head of mp->to  */
						ap = alias_list;
						while (ap != NULL)
							{
							struct list *ap_next = ap->next;
							
							ap->next = mp->to;
							mp->to = ap;
							ap = ap_next;
							}
						}

					usprintf (mp->s, Ok, original_envelope);
					}

				} /* if (LOCAL address) */
			else if (bad_sender != NULL || !smtp_relay)
				{
				usprintf (mp->s, NoRelay, cp);
				smtplog ("relay attempt rejected: To: %s From: %s", cp, mp->from);
				if (smtpverbose)
					tprintf ("SMTP: Rejecting relay to %s from %s\n", cp, mp->from);
				}
			else
				{
				/* a remote address is added to the list */
				addlist (&mp->to, cp, address_type);
				usprintf (mp->s, WillForward, cp);
				}

			/*  GeK - stash first valid recipient envelope for header  */
			if (mp->to != NULL && mp->envelope_to == NULL)
				mp->envelope_to = original_envelope;
			else
				free (original_envelope);

			break;
		case HELP_CMD:
			usputs (mp->s, Help);
			break;
		case DATA_CMD:
			if (mp->to == NULL)
				usputs (mp->s, Needrcpt);
			else if ((mp->data = tmpfile ()) == NULLFILE)
				usputs (mp->s, Ioerr);
			else
				{
				/*  Accept message, getmsgtxt() also sends acceptance  */
				if (getmsgtxt (mp) == 0)
					messages_received++;
				}
			break;
		case RSET_CMD:
			del_list (mp->to);
			mp->to = NULLLIST;
			free (mp->envelope_to);		/*  GeK - reset first recipient  */
			mp->envelope_to = NULL;
			usputs (mp->s, Reset);
			break;
		case EXPN_CMD:
		case VRFY_CMD:	/*  GeK - treat EXPN and VRFY almost the same  */
			if (*arg == '\0')
				{
				usputs (mp->s, Syntax);
				break;
				}

			/* rewrite address if possible */
			if ((newaddr = rewrite_address (arg)) != NULL)
				{
				if (stricmp (newaddr, arg) != 0)
					{
					strcpy (buf, newaddr);
					arg = buf;
					}
				free (newaddr);
				}

			address_type = validate_address (arg);
			if (address_type == BADADDR)
				{
				usprintf (mp->s, Unknown, arg);
				break;
				}
			else if (address_type == DOMAIN)
				{
				usprintf (mp->s, NoRelay, arg);
				break;
				}

			alias_list = expand_local_alias (arg);
			if (alias_list == NULL)
				{
				usprintf (mp->s, Noalias, arg);
				break;
				}
			else if ((cmdp - commands) == VRFY_CMD)
				{
				usprintf (mp->s, Ok, arg);
				}
			else	/*  EXPN  */
				{
				ap = alias_list;
				while (ap != NULL)
					{
					int	has_domain = (strchr (ap->val, '@') != NULL);

					usprintf (mp->s, "250%c<%s%s%s>\n",
					          ap->next != NULL ? '-' : ' ',
					          ap->val,
					          has_domain ? "" : "@",
					          has_domain ? "" : Hostname);
					ap = ap->next;
					}
				}

			del_list (alias_list);
			break;			
		}	/* switch (cmdp - commands) */

	goto loop;

quit:
	/*  GeK - added message count stats to close log  */
	log (mp->s, "close SMTP: %u/%u messages", messages_received, mail_from_count);
	close_s (mp->s);
	mail_clean (mp);
	free (bad_sender);		/*  GeK  */
	smtptick (NULL);							 /* start SMTP daemon
												  * immediately */
	}

/* read the message text */
static int
    getmsgtxt (mp)
struct smtpsv *mp;
	{
	char buf[LINELEN];
	register char *p = buf;
	time_t t;
	int check_dot = 1;			/*  GeK - moved from loop - check for "."  */
	long msgid;

	/*  GeK - remember ID for this message  */
	msgid = get_msgid();

#ifdef	VERIFY_FROM_ADDRESS
	if (!mp->from_verified)
		fprintf (mp->data, "X-Warning: %s: <%s> failed DNS check\n", Hostname, mp->from);
#endif	/* VERIFY_FROM_ADDRESS */

	/* Add timestamp; ptime adds newline */
	time (&t);
	fprintf (mp->data, "Received: from %s ([%s]) by %s\n",
	         mp->system != NULL ? mp->system : "(no HELO)",
	         inet_ntoa (mp->sender_ip_addr),
	         Hostname);
	fprintf (mp->data, "          with %sSMTP id AA%ld", mp->esmtp ? "E" : "", msgid);
	/*  Check for original envelope and only one recipient  */
	if (mp->envelope_to != NULL && mp->to->next == NULL)
		fprintf (mp->data, " for <%s>;\n          ", mp->envelope_to);
	else
		fputs ("; ", mp->data);
	fputs (ptime (&t), mp->data);

	if (ferror (mp->data))
		{
		usputs (mp->s, Ioerr);
		return 1;
		}
	else
		{
		usputs (mp->s, Enter);
		}

	while (1)
		{

		p = buf;						/* GT 11 Dec 93						*/
		if (recvline (mp->s, p, sizeof (buf)) == -1)
			{
			return 1;
			}

		/* If the line starts with a '.', this is either the end of the
		 * message or a line where a dot has been added for "dot
		 * transparency"
		 */
		if (check_dot && *p == '.')
			{
			/* Strip off leading '.'; if there is nothing else on the
			 * line, this is the end of the data.  Otherwise, we have
			 * managed to remove the protecting dot and p now points at
			 * the unprotected message line.
			 */
			if (*++p == '\n')	/*  GeK - check \n not \0  */
				{
				int rc;					/* result code						*/
				
				/* Also sends appropriate response */
				
				/*  GeK - if error on mp->data, don't attempt mailit  */
				if (ferror (mp->data))
					rc = 1;
				else
					rc = mailit (mp->data, mp->from, mp->to);
				if (fclose (mp->data))
					rc = 1;

#ifdef	COMMIT_CACHE
				/*  GeK - ensure data really committed to disk before ACK'ing  */
				if (commit_cache && rc == 0)
					{
					pwait (NULL);	/*  Allow any waiting processes to go first  */
					smartdrive_commit_cache ();
					pwait (NULL);
					}
#endif	/* COMMIT_CACHE */

				mp->data = NULLFILE;
				if (rc != 0)
					usputs (mp->s, Ioerr);
				else
					/*  GeK - added ID  */
					usprintf (mp->s, Sent, msgid);

#ifdef	notdef	/*  GeK - don't reset to list - now allows very broken DATA DATA behaviour  */
				del_list (mp->to);
				mp->to = NULLLIST;
#endif	/* notdef */
				return rc;
				}

			}	/* if (check_dot && *p == '.') */

		/*  If buffer has newline then check for dot next time around  */
		check_dot = (strchr (p, '\n') != NULL);

#ifdef	notdef	/*  GeK - don't escape here, done on copy to mailbox  */
		/* for UNIX mail compatiblity */

		if (strncmp (p, "From ", 5) == 0)
			(void) fputc ('>', mp->data);
#endif

		/* Append to data file */

		/* 
		 *  GeK - removed error check and return here to avoid screwing up
		 *  state machine by not completeing reception of entire message
		 */
		fputs (p, mp->data);
				
		/*
		 *  GeK - this really shouldn't be needed but it's added purposely
		 *  to split long (>255 chr) lines to avoid breaking Ted.com
		 */
#ifdef	BREAK_LONG_LINES
		if (check_dot == 0)
			fputc ('\n', mp->data);
#endif	/* BREAK_LONG_LINES */

		}
	}

/* Create control block, initialize */
static struct smtpsv *
       mail_create ()
	{
	register struct smtpsv *mp;

	mp = (struct smtpsv *) callocw (1, sizeof (struct smtpsv));
	mp->from = strdup ("");						 /* Default to null From
												  * address */
	return mp;
	}

/* Free resources, delete control block */
static void
     mail_clean (mp)
register struct smtpsv *mp;
	{
	if (mp == NULLSMTPSV)
		return;

	free (mp->system);
	free (mp->from);
	if (mp->data != NULLFILE)
		fclose (mp->data);

	free (mp->envelope_to);		/*  GeK - free additionals  */
	del_list (mp->to);
	free ((char *) mp);
	}

/* Given a string of the form <user@host>, extract the part inside the
 * brackets and return a pointer to it.
 */
char *
     getname (cp)
register char *cp;
	{
	register char *cp1;

	if ((cp = strchr (cp, '<')) == NULLCHAR)
		return NULLCHAR;

	cp++;										 /* cp -> first char of name */
	if ((cp1 = strchr (cp, '>')) == NULLCHAR)
		return NULLCHAR;

	*cp1 = '\0';
	return cp;
	}

#ifdef NOVELL

long int get_mail_id (char *user, char *server)
	{
	int connectionID = 1;
	int cCode = 0;
	int DefCon;
	long int retval;
	int err;

	cCode = GetConnectionID (server, &connectionID);

	DefCon = GetDefaultConnectionID ();

	if (DefCon != connectionID)
		{
		SetPreferredConnectionID (connectionID);
		}

	cCode = GetBinderyObjectID (user, OT_USER, &retval);

	if (cCode != 0)
		{
		retval = 0;
		}

	if (DefCon != connectionID)
		{
		SetPrimaryConnectionID (DefCon);
		}

	return retval;
	}

void send_alert (char *user, char *server, char *from)
	{
	int numcon, conlist[100], reslist[100];
	int connectionID = 1;
	int cCode = 0;
	int DefCon;
	long int retval;
	int err;
	char message[128];

	cCode = GetConnectionID (server, &connectionID);

	DefCon = GetDefaultConnectionID ();

	if (DefCon != connectionID)
		{
		SetPreferredConnectionID (connectionID);
		}

	cCode = GetObjectConnectionNumbers (user, OT_USER, &numcon, conlist, 100);

	sprintf (message, "You have mail from %s", from);

	message[55] = 0;							 /* Nasty truncation to cope
												  * with bindings */

	cCode = SendBroadcastMessage (message, conlist, reslist, numcon);

	if (DefCon != connectionID)
		{
		SetPrimaryConnectionID (DefCon);
		}

	}

#endif

/*
** copy_data
**
** This function copies all available data from an input file to an
** output file.  This is done in chunks of LINELEN for efficiency.
** In addition, the function deschedules itself every so often so
** that large copy operations of this type do not completely block
** processing of other operations.
*/
static void
copy_data (FILE *from, FILE *to)
{
	char *buf;
	int	last_had_newline = TRUE;

	buf = mallocw (LINELEN);
	while (fgets (buf, LINELEN, from) != NULL)
		{
		/*  Escape 'From ' at the beginning of a line for unix mail compatibility  */
		if (last_had_newline && strncmp (buf, "From ", 5) == 0)
			fputc ('>', to);
		if (fputs (buf, to) == EOF)
			break;
		pwait (NULL);
		last_had_newline = (strchr (buf, '\n') != NULL);
		}
	free (buf);
}

/* General mailit function. It takes a list of addresses which have already
** been verified and expanded for aliases. Base on the current mode the message
** is place in an mbox, the outbound smtp queue or the rqueue interface
*/
static int
    mailit (data, from, tolist)
FILE *data;
char *from;
struct list *tolist;
	{
	struct list *ap, *dlist = NULLLIST;
	register FILE *fp;
	char mailbox[85], *cp, *host, *qhost;
	int fail = 0;
	time_t t;
	int32 msgid;	/*  GeK - moved from below for tracking queuejob fails  */
	int	retry;		/*  GeK - mailbox lock retry interval (ms)  */

	if ((Smtpmode & QUEUE) != 0)
		return (router_queue (data, from, tolist));

	do
		{
		qhost = NULLCHAR;
		for (ap = tolist; ap != NULLLIST; ap = ap->next)
			{
#if	defined (DEBUG_LOCKS)
			smtplog ("mailit (): ap->val = %s", ap->val);
			smtplog ("mailit (): ap->type = %d", ap->type);
#endif
			if (ap->type == DOMAIN)
				{
				if ((host = strrchr (ap->val, '@')) != NULLCHAR)
					host++;
				else
					host = Hostname;

				if (qhost == NULLCHAR)
					qhost = host;

				if (stricmp (qhost, host) == 0)
					{
					ap->type = BADADDR;
					addlist (&dlist, ap->val, 0);
					}

				}

			}
		if (qhost != NULLCHAR)
			{
#if	defined (DEBUG_LOCKS)
			smtplog ("mailit (): qhost = %s, queueing", qhost);
#endif
			rewind (data);
			fail = queuejob (data, qhost, dlist, from, &msgid);
			del_list (dlist);
			dlist = NULLLIST;
			if (fail)
				{
				if (smtpverbose)
					tprintf ("Mail failed: queue job %ld host %s from %s\n", msgid, qhost, from);
				smtplog ("queue job %ld failed: Host: %s From: %s", msgid, qhost, from);
				return (fail);
				}
			}

		}
	while (qhost != NULLCHAR)

	;

	for (ap = tolist; ap != NULLLIST; ap = ap->next)
		{
		/*  GeK - for long lists of non-local, we could hog CPU here?  */
		pwait (NULL);

		if (ap->type != LOCAL)
			{
			ap->type = DOMAIN;
			continue;
			}

		rewind (data);
		/* strip off host name of LOCAL addresses */
		if ((cp = strchr (ap->val, '@')) != NULLCHAR)
			*cp = '\0';

		/* truncate long user names */

		if (strlen (ap->val) > MBOXLEN)
			ap->val[MBOXLEN] = '\0';

		/* if mail file is busy save it in our smtp queue and let the smtp
		 * daemon try later. */

#ifdef NOVELL

		if (!novell_start)
			{
#endif
			retry = 125;
			while (retry < 1000
			       && mlock (Mailspool, ap->val))
				{
				pause (retry);
				retry <<= 1;
				}

			if (retry >= 1000)
				{
#if	defined (DEBUG_LOCKS)
				smtplog ("mailit (): mail spool file %s, %s locked, queueing",
						 Mailspool, ap->val);
#endif
				addlist (&dlist, ap->val, 0);
				fail = queuejob (data, Hostname, dlist, from, &msgid);
				del_list (dlist);
				dlist = NULLLIST;
				if (!fail)
					{
					/*  GeK - delay_job now takes 8 char prefix rather than int32 id  */
					char prefix_buf[16];

					sprintf (prefix_buf, "%ld", msgid);
					delay_job (prefix_buf);
					}
				else
					{
					if (smtpverbose)
						tprintf ("Mail failed: queue job %ld host %s from %s\n", msgid, Hostname, from);
					smtplog ("queue job %ld failed: Host: %s From: %s", msgid, Hostname, from);
					return (fail);
					}
				}
#ifdef NOVELL

			}
#endif
		else
			{
			char *buf;
			int tocnt = 0;
			int last_had_newline = TRUE;	/*  GeK - detect start of long lines  */

#ifdef NOVELL
			long int oid;
			char box[13];
			struct ffblk fblock;
			static int done_rand = 0;

			if (novell_start)
				{
				oid = get_mail_id (ap->val, novell_server_name);

				ltoa (oid, box, 16);

				if (!done_rand)
					{
					randomize ();				 /* This is non-deterministic */
					done_rand = 1;
					}

				do
					{
					sprintf (mailbox, "%s/%s/%s/%04x%04x.%s", novell_server_name, "SYS:/MAIL", box, rand (), rand (), novell_mail_ext);
					}
				while (!findfirst (mailbox, &fblock, 0))

				;
				}
			else
#endif
				{
				sprintf (mailbox, "%s/%s.txt", Mailspool, ap->val);
				}

#ifndef	AMIGA
			if ((fp = fopen (mailbox, APPEND_TEXT)) != NULLFILE)
				{
#	ifdef NOVELL
				if (novell_start)
					{
					send_alert (ap->val, novell_server_name, from);
					}
#	endif

#else
			if ((fp = fopen (mailbox, "r+")) != NULLFILE)
				{
				(void) fseek (fp, 0L, 2);
#endif
				buf = mallocw (LINELEN);
				time (&t);
				if (smtp_separator)
					fprintf (fp, "%s\n", smtp_separator);

				fprintf (fp, "From %s %s", from, ctime (&t));
				host = NULLCHAR;
				/*
				 * SCT - Write message header ourselves, so we can look for
				 * "To:" and "Cc:" lines and add "Apparently-To:" to the end
				 * if we don't see one; also optionally escape "Status:"
				 * lines in the header, which shouldn't be present in new
				 * mail
				 */
				while (fgets (buf, LINELEN, data) != NULLCHAR)
					{
					int	has_newline;
					
					has_newline = (strchr (buf, '\n') != NULL);

					if (last_had_newline && buf[0] == '\n')
						break;	/*  while()  */

#ifdef	ESCAPE_STATUS_HDR
					if (last_had_newline &&
					    strnicmp(Hdrs[STATUS], buf, strlen(Hdrs[STATUS]) - 1) == 0)
						fputc('>', fp);
#endif
					if (fputs (buf, fp) == EOF)
						break;
					if (last_had_newline)
						{
						rip (buf);
						switch (htype (buf))
							{
							case TO:
							case CC:
								++tocnt;
								break;
							case RRECEIPT:
								if (smtp_receipt
								    && (cp = getaddress (buf, 0)) != NULL)
									{
									free (host);
									host = strdup (cp);
									}

								break;
							}
						}
					last_had_newline = has_newline;
					/*  GeK - assume write slow, give up some time  */
					pwait (NULL);
					}

				free (buf);

				/* 
				 *  GeK - this was in above loop but would never get executed
				 *  in the unlikely event of no message body at all ("DATA\n.\n")
				 */
				if (tocnt == 0)
					{
					fprintf (fp, "%s%s", Hdrs[APPARTO], ap->val);
					if (strchr (ap->val, '@') == NULL)
						fprintf (fp, "@%s", Hostname);
					fputc ('\n', fp);
					}
				fputc ('\n', fp);

				copy_data (data, fp);

#ifdef	BLANK_LINE_BETWEEN_MSGS
				/* Leave a blank line between msgs */
				/*  GeK - does anyone need a blank line between msgs? PCElm works OK without  */
				fputc ('\n', fp);
#endif	/* BLANK_LINE_BETWEEN_MSGS */

				if (ferror (fp))
					fail = 1;

				if (fclose (fp))
					fail = 1;	/*  GeK - catch poss. close error  */

				/*  GeK - mail notification moved below  */

				if (host != NULLCHAR)
					{
					rewind (data);				 /* Send return receipt */
					mdaemon (data, host, NULLLIST, 0);
					free (host);
					}

				}
			else
				fail = 1;

#ifdef NOVELL
			if (!novell_start)
				{
				(void) rmlock (Mailspool, ap->val);
				}
#else
			(void) rmlock (Mailspool, ap->val);
#endif
			if (fail)
				{
				if (smtpverbose)
					tprintf ("Mail failed: to %s from %s\n", ap->val, from);
				smtplog ("deliver failed: To: %s From: %s", ap->val, from);
				break;	/*  for (tolist)  */
				}

			/*  GeK - mail notification moved here to prevent false alerts  */
			if (smtpverbose)
				tprintf ("New mail arrived for %s from %s\n", ap->val, from);

			if (Smtpbeep)
				printf ("\a\a");

			smtplog ("deliver: To: %s From: %s", ap->val, from);
			}
		}

	return fail;
	}

/* Return Date/Time in Arpanet format in passed string */
/*  GeK - modified to use Timezone offset instead  */
char *
ptime (time_t *t)
	{
	register struct tm *ltm;	/*  Ptr to localtime volatile struct  */
	int                 tzdiff;	/*  Timezone difference in minutes  */
	static char         str[48];

	/* Read the local system time */
	ltm = localtime (t);

	/*  difference localtime from GMT in minutes  */
	tzdiff = (int) (-timezone/60) + (ltm->tm_isdst ? 60 : 0);

/*  Macro to convert +tzdiff to +HHMM  */
#define TZDIFF_AS_HHMM(d) ((100 * ((d) / 60)) + ((d) % 60))

	/*  GeK - RFC1123 - 2 digit day, 4 digit year, TZ diff from UTC  */
	sprintf (str, "%s, %02u %s %04u %02u:%02u:%02u %+05d (%.5s)\n",
	         Days [ltm->tm_wday],
	         ltm->tm_mday,
	         Months [ltm->tm_mon],
	         ltm->tm_year + 1900,	/*  GeK - +1900  */
	         ltm->tm_hour,
	         ltm->tm_min,
	         ltm->tm_sec,
	         TZDIFF_AS_HHMM (tzdiff),
	         tzname [ltm->tm_isdst ? 1 : 0 ]
	        );

	return (str);
	}

long
     get_msgid ()
	{
	char sfilename[80];
	char s[20];
	register long sequence = 0;
	FILE *sfile;
	int	retry = 125;	/*  Retry interval in ms  */

	while (mlock (Mailqdir, "sequence"))
		{
		if (retry > 2000)
			{
			/*  We're in trouble, try for a unique-ish value ...  */
			return (((time(NULL) / 1000L) * 1000L + (msclock() % 1000L)) % 100000000L);
			}

		pause (retry);
		retry <<= 1;
		}

	sprintf (sfilename, "%s/sequence.seq", Mailqdir);
	sfile = fopen (sfilename, READ_TEXT);

	/* if sequence file exists, get the value, otherwise set it */
	if (sfile != NULL)
		{
		(void) fgets (s, sizeof (s), sfile);
		sequence = atol (s);
		/* Keep it in range of and 8 digit number to use for dos name
		 * prefix. */
		/* GeK - changed > to >= to prevent overflow below  */
		if (sequence < 0L || sequence >= 99999999L)
			sequence = 0;

		fclose (sfile);
		}

	/* increment sequence number, and write to sequence file */

	sfile = fopen (sfilename, WRITE_TEXT);
	fprintf (sfile, "%ld", ++sequence);
	fclose (sfile);
	rmlock (Mailqdir, "sequence");
	return sequence;
	}

#ifdef	MSDOS
/* Illegal characters in a DOS filename */	/*  GeK - added a few more  */
static char baddoschars[] = "\"[]:|<>+=;,. \t\\/*?";
/*  GeK - MS-DOS[tm] device names to check for  */
static char dos_devices[] = "auxconnulprn";
#endif

/* test if mail address is valid */
int
    validate_address (s)
char *s;
	{
	char *cp, *t;
	int32 addr;

	/* if address has @ in it the check dest address */
	if (*s == '@' && (cp = strrchr (s, ':')) != NULLCHAR)
		for (t = s, cp++; (*t++ = *cp++) != 0;)
			;

	if ((cp = strrchr (s, '@')) != NULLCHAR)
		{
		cp++;
		/* 1st check if its our hostname if not then check the hosts file
		 * and see if we can resolve ther address to a know site or one of
		 * our aliases */
		if (stricmp (cp, Hostname) != 0)	/*  GeK - case insensitive  */
			{
			if ((addr = mailroute (cp, 0)) == 0
				&& (Smtpmode & QUEUE) == 0)
				return BADADDR;

			if (ismyaddr (addr) == NULLIF)
				return DOMAIN;

			}

		/* on a local address remove the host name part */

		*--cp = '\0';
		}

	/* if using an external router leave address alone */

	if ((Smtpmode & QUEUE) != 0)
		return LOCAL;

	/* check for the user%host hack */

	if ((cp = strrchr (s, '%')) != NULLCHAR)
		{
		*cp = '@';
		cp++;
		/* reroute based on host name following the % seperator */
		if (mailroute (cp, 0) == 0)
			return BADADDR;
		else
			return DOMAIN;
		}

	/*  GeK - drop completely empty address  */
	if (*s == '\0')
		return BADADDR;

#ifdef MSDOS									 /* dos file name checks */
	/* Check for characters illegal in MS-DOS file names */

	for (cp = baddoschars; *cp != '\0'; cp++)
		{
		if (strchr (s, *cp) != NULLCHAR)
			return BADADDR;

		}

	/*  GeK - check for standard DOS device names (colon caught above)  */
	for (cp = dos_devices; *cp != '\0'; cp += 3)
		{
		if (strnicmp (s, cp, 3) == 0 && s[3] == '\0')
			return BADADDR;
		}
	if (((strnicmp (s, "lpt", 3) == 0 || strnicmp (s, "com", 3) == 0)
	     && isdigit (s[3]) && s[4] == '\0')
	    || stricmp (s, "clock$") == 0)
		return BADADDR;
#endif

	return LOCAL;
	}

/* place a mail job in the outbound queue */
int
    queuejob (dfile, host, to, from, msgid)
FILE *dfile;
char *host;
struct list *to;
char *from;
int32 *msgid;
	{
	FILE *fp;
	struct list *ap;
	char tmpstring[80], prefix[9], *buf;
	register int cnt;
	int32 this_msgid;
	int fail = 0;

	this_msgid = get_msgid ();
	sprintf (prefix, "%ld", this_msgid);
	if (msgid)
		*msgid = this_msgid;

	mlock (Mailqdir, prefix);
	sprintf (tmpstring, "%s/%s.txt", Mailqdir, prefix);
	if ((fp = fopen (tmpstring, WRITE_TEXT)) == NULLFILE)
		{
		fail = 1;
		goto queuejob_quit;
		}

	buf = mallocw (LINELEN);
	while ((cnt = fread (buf, 1, LINELEN, dfile)) > 0)
		{
		if (fwrite (buf, 1, cnt, fp) != cnt)
			break;
		/*  GeK - assume write slow, give up some time  */
		pwait (NULL);
		}
	free (buf);

	if (ferror (fp))
		fail = 1;

	if (fclose (fp))
		fail = 1;

	if (fail)
		goto queuejob_quit;

	sprintf (tmpstring, "%s/%s.wrk", Mailqdir, prefix);
	if ((fp = fopen (tmpstring, WRITE_TEXT)) == NULLFILE)
		{
		fail = 1;
		goto queuejob_quit;
		}

	fprintf (fp, "%s\n%s\n", host, from);
	for (ap = to; ap != NULLLIST; ap = ap->next)
		{
#ifdef	PERCENT_HACK_ROUTE_FIX
		/*  GeK - workaround for post.demon's inability to reliably maintain the %-hack */

		char	*host_ptr;
		
		if (strchr (ap->val, '%') != NULL					/*  Has % hack  */
		    && smtp_gateway != NULL							/*  SMTP gateway defined  */
		    && (host_ptr = strrchr (ap->val, '@')) != NULL	/*  Has '@' */
		    && stricmp (host_ptr+1, host) == 0				/*  Matches queue host  */
		    && stricmp (smtp_gateway, host) == 0)			/*  SMTP gateway == queue host  */
			{
			*host_ptr = '\0';						/*  Trim off queue host  */
			if (strchr (ap->val, '@') == NULL		/*  Check only one '@'  */
			    && (host_ptr = strrchr (ap->val, '%')) != NULL)		/*  Find last '%' (shouldn't fail)  */
				{
				*host_ptr = '@';						/*  Convert to mail domain  */
				}
			}

#endif	/*  PERCENT_HACK_ROUTE_FIX  */

		fprintf (fp, "%s\n", ap->val);
		/*  GeK - assume write slow, give up some time  */
		pwait (NULL);
#ifdef	notdef
		smtplog ("queue job %s To: %s From: %s", prefix, ap->val, from);
#endif
		}

	if (ferror (fp))
		fail = 1;
	if (fclose (fp))
		fail = 1;

	/*  GeK - only one queue log entry  */
	if (!fail)
		smtplog ("queue job %s To: %s %sFrom: %s", prefix, to->val, (to->next != NULL) ? "... " : "", from);

queuejob_quit:
	rmlock (Mailqdir, prefix);
	return (fail);
	}

/* Deliver mail to the appropriate mail boxes */
static int
    router_queue (data, from, to)
FILE *data;
char *from;
struct list *to;
	{
	register struct list *ap;
	FILE *fp;
	char tmpstring[80];
	char prefix[9];
	int fail = 0;

	sprintf (prefix, "%ld", get_msgid ());
	mlock (Routeqdir, prefix);
	sprintf (tmpstring, "%s/%s.txt", Routeqdir, prefix);
	if ((fp = fopen (tmpstring, WRITE_TEXT)) == NULLFILE)
		{
		fail = 1;
		goto router_queue_quit;
		}

	rewind (data);
	copy_data (data, fp);
	if (ferror (fp))
		fail = 1;

	if (fclose (fp))
		fail = 1;

	if (fail)
		goto router_queue_quit;

	sprintf (tmpstring, "%s/%s.wrk", Routeqdir, prefix);
	if ((fp = fopen (tmpstring, WRITE_TEXT)) == NULLFILE)
		{
		fail = 1;
		goto router_queue_quit;
		}

	fprintf (fp, "From: %s\n", from);
	for (ap = to; ap != NULLLIST; ap = ap->next)
		{
		fprintf (fp, "To: %s\n", ap->val);
		}

	if (ferror (fp))
		fail = 1;
	if (fclose (fp))
		fail = 1;
	if (fail)
		goto router_queue_quit;

	smtplog ("rqueue job %s From: %s", prefix, from);

router_queue_quit:
	(void) rmlock (Routeqdir, prefix);
	return (fail);
	}

/* add an element to the front of the list pointed to by head */
struct list *
addlist (struct list **head, char *val, int type)
{
	struct list *tp;

	tp = (struct list *) callocw (1, sizeof (struct list));

	/* allocate storage for the char string */
	tp->val = strdup (val);
	tp->type = type;

	tp->next = *head;
	*head = tp;

	return tp;
}


#ifdef	notdef	/*  GeK - no longer needed  */
/* Remove and entry from the list ...			CMS 19/12/92
	return TRUE on success or FALSE on failure */
static int remlist (struct list ** head, struct list * unwanted)
	{
	struct list *ptr, **prev;

	prev = head;
	for (ptr = *head; ptr; ptr = ptr->next)
		if (ptr == unwanted)
			{
			*prev = ptr->next;
			free (ptr->val);	/*  GeK - plug memory leak  */
			free (ptr);
			break;
			}
		else
			prev = &ptr->next;

	return ptr == unwanted;
	}
#endif	/* notdef */

/*
 *  GeK - moved from RCPT TO, this function expands the given address into
 *        a returned list of valid addresses.  NULL list indicates failure.
 */
static struct list *
expand_local_alias (char *address)
{
	failure default_fail = NO_FAIL;
	struct list	*alias_list = NULL;
	struct list	*last_ap = NULL, *ap, *next_ap;

	/*  Expand address into alias list  */
	expandalias (&alias_list, address, &default_fail);

	for (ap = alias_list; ap != NULL; last_ap = ap, ap = next_ap)
		{
		next_ap = ap->next;

		if (ap->type == LOCAL
			&& default_fail != FAIL_DELIVER
			&& validate_user (ap->val) == 0)
			{
			/*  expandalias now handles default expansion, so any failure
			 *  here means default alias expansion also failed
			 */
			/*  Log alias expansion failures  */
			if (default_fail == FAIL_DEFAULT || stricmp (address, ap->val) != 0)
				log (-1, "Mail alias '%s' has invalid %suser '%s'",
				     address,
				     default_fail == FAIL_DEFAULT ? "default " : "",
				     ap->val);

			/*  Remove entry from list and free mem, adjust ap for for() loop  */
			free (ap->val);
			free (ap);
			if (last_ap == NULL)
				ap = alias_list = next_ap;
			else
				ap = last_ap->next = next_ap;
			}
		}
	
	return (alias_list);
}

#define SKIPWORD(X) while(*X && *X!=' ' && *X!='\t' && *X!='\n' && *X!= ',') X++;
#define SKIPSPACE(X) while(*X ==' ' || *X =='\t' || *X =='\n' || *X == ',') X++;

/* check for and alias and expand alias into an address list */
/*
 *  GeK - substantially rewritten (again)
 *        Returns: head contains expanded alias
 *                 *default_fail updated:
 *                    No Alias file: FAIL_DELIVER
 *                    Alias file with default deliver: FAIL_DELIVER
 *                    Alias file with default mailbox: FAIL_DEFAULT, list contains default expansion
 */
static struct list *
expandalias (struct list ** head,		/*  list to expand alias into  */
             char *user,				/*  alias to expand  */
             failure *default_fail)		/*  Should pt to NO_FAIL initially  */
{
	FILE *fp;
	char *s, *p;
	int inalias = 0;
	struct list *tp;
	char *buf;
	char *search_alias;				/*  alias we're looking to expand  */
	char *default_alias = NULL;		/*  default mailbox  */
	int pass = 0;

	/* no alias file found */
	if ((fp = fopen (Alias, READ_TEXT)) == NULLFILE)
		{
		*default_fail = FAIL_DELIVER;

		/*  GeK - removed obselete MAILB MB, MG, MR record searching  */

		/*  Add local address entry  */
		return addlist (head, user, LOCAL);
		}

	buf = mallocw (LINELEN);
	search_alias = user;

search_alias_file:
	while (fgets (buf, LINELEN, fp) != NULLCHAR)
		{
		p = buf;
		if (*p == '#' || *p == '\n')	/*  GeK - fgets can't put '\0' in buf[0]  */
			continue;

		rip (p);

		/* if not in an matching entry skip continuation lines */
		if (!inalias && isspace (*p))
			continue;

		/* when processing an active alias check for a continuation */

		if (inalias)
			{
			if (!isspace (*p))
				break;							 /* done */
			}
		else
			{
			s = p;
			SKIPWORD (p);
			if (*p != '\0')		/*  GeK - else p runs off into oblivion ...  */
				*p++ = '\0';					 /* end the alias name */
			SKIPSPACE (p);
			if (stricmp (s, "default") == 0
			    && *default_fail == NO_FAIL)
				{
				/*  GeK - trim whitespace and ensure mailbox is terminated  */
				s = p;
				SKIPWORD (p);
				*p = '\0';
				if (stricmp (s, "deliver") == 0)
					*default_fail = FAIL_DELIVER;
				else if (*s != '\0')
					{
					*default_fail = FAIL_DEFAULT;
					default_alias = strdup (s);
					}
				/*  GeK - The line 'default default' might do strange things here  */
				continue;
				}

#ifdef	WILDCARD_ALIAS	/*  GeK - good idea or not?  */
			if (!wildmat (search_alias, s, NULL))
#else
			if (stricmp (s, search_alias) != 0)		/*  GeK - case insensitive match  */
#endif
				continue;						 /* no match go on */

			inalias = 1;
			}

		/* process the recipients on the alias line */
		SKIPSPACE (p);
		while (*p != '\0' && *p != '#')
			{
			s = p;
			SKIPWORD (p);
			if (*p != '\0')
				*p++ = '\0';

			/* find hostname */

			if (strchr (s, '@') != NULLCHAR)
				tp = addlist (head, s, DOMAIN);
			else
#ifdef	RECURSIVE_ALIAS		/*  GeK from todo 12/93 - attributed to OGS  */
				/*  GeK - added sanity checks to try to prevent loops  */
				if (stricmp (s, search_alias) != 0
#  ifdef	WILDCARD_ALIAS
				    && !wildmat (search_alias, s, NULL)
#  endif
				   )
					tp = expandalias (head, s, NULL, NULL);
				else
					tp = addlist (head, s, LOCAL);
#else	/* !RECURSIVE_ALIAS */
				tp = addlist (head, s, LOCAL);
#endif

			SKIPSPACE (p);
			}
		}

	/*  GeK - 'default mailbox' found and no match for user  */
	if (!inalias
	    && *default_fail == FAIL_DEFAULT
	    && pass++ == 0)
		{
		/*  Search again to expand default alias  */
		search_alias = default_alias;
		rewind (fp);
		goto search_alias_file;
		}

	fclose (fp);
	free (buf);

	if (!inalias)		/* no alias found treat as a local address */
		tp = addlist (head, search_alias, LOCAL);

	free (default_alias);

	return tp;
	}

#if	defined(ANSIPROTO)
static void
     smtplog (char *fmt,...)
	{
	va_list ap;
	char *cp;
	long t;
	FILE *fp;

	if ((fp = fopen (Maillog, APPEND_TEXT)) == NULLFILE)
		return;

	time (&t);
	cp = ctime (&t);
	rip (cp);
	fprintf (fp, "%s ", cp);
	va_start (ap, fmt);
	vfprintf (fp, fmt, ap);
	va_end (ap);
	fputc ('\n', fp);
	fclose (fp);
	}

#else

static void
     smtplog (fmt, arg1, arg2, arg3, arg4)
char *fmt;
int arg1, arg2, arg3, arg4;
	{
	char *cp;
	long t;
	FILE *fp;

	if ((fp = fopen (Maillog, APPEND_TEXT)) == NULLFILE)
		return;

	time (&t);
	cp = ctime (&t);
	rip (cp);
	fprintf (fp, "%s ", cp);
	fprintf (fp, fmt, arg1, arg2, arg3, arg4);
	fprintf (fp, "\n");
	fclose (fp);
	}

#endif

/* send mail to a single user. Can be called from the ax24 mailbox or
** from the return mail function in the smtp client
*/
static int
mailuser (FILE *data, char *from, char *to)
{
	int address_type, ret;
	struct list *tolist = NULL;
	char	*rewritten_to;

	/*
	 *  GeK - rewrite address first.  For this to work,
	 *        rewrite file MUST NOT rewrite all mail locally
	 */
	rewritten_to = rewrite_address (to);
	if (rewritten_to != NULL)
		to = rewritten_to;

	/* check if address is ok */
	if ((address_type = validate_address (to)) == BADADDR)
		{
		log (-1, "mailuser: bad address %s", to);
		free (rewritten_to);
		return 1;
		}
	/* if a local address check for an alias */

	if (address_type == LOCAL)
		{
		tolist = expand_local_alias (to);

		if (tolist == NULL)
			{
			log (-1, "mailuser: bad address %s", to);
			free (rewritten_to);
			return 1;
			}
		}
	else
		{
		/* a remote address is added to the list */
		addlist (&tolist, to, address_type);
		}

	ret = mailit (data, from, tolist);
	del_list (tolist);
	free (rewritten_to);
	return ret;
}

/* Mailer daemon return mail mechanism */
int
mdaemon (FILE *data,				 /* pointer to rewound data file */
         char *to,					 /* Overridden by Errors-To: line if bounce is
         							  * true */
         struct list *lp,			 /* error log for failed mail */
         int bounce)				 /* True for failed mail, otherwise return
         							  * receipt */
{
	time_t t;
	FILE *tfile;
	char *buf, *cp, *newto = NULLCHAR;
	int cnt;
	int last_had_newline;
	int file_error = FALSE;	/*  GeK  */

	buf = mallocw (LINELEN);
	if (to == NULLCHAR || (to != NULLCHAR && *to == '\0') || bounce)
		{
		last_had_newline = TRUE;
		while (fgets (buf, LINELEN, data) != NULLCHAR)
			{
			if (last_had_newline && buf[0] == '\n')
				break;

			/* Look for Errors-To: */

			if (htype (buf) == ERRORSTO &&
				(cp = getaddress (buf, 0)) != NULLCHAR)
				{
				free (newto);
				newto = strdup (cp);
				break;
				}

			last_had_newline = (strchr (buf, '\n') != NULL);
			}

		if (newto == NULLCHAR && ((to != NULLCHAR && *to == '\0') ||
								  to == NULLCHAR))
			{
			free (buf);
			return -1;
			}

		rewind (data);
		}

	if ((tfile = tmpfile ()) == NULLFILE)
		{
		free (buf);
		return -1;
		}

	time (&t);
	fprintf (tfile, "%s%s", Hdrs[DATE], ptime (&t));
	fprintf (tfile, "%s<%ld@%s>\n", Hdrs[MSGID], get_msgid (), Hostname);
	fprintf (tfile, "%sMail Delivery Subsystem <MAILER-DAEMON@%s>\n",
			 Hdrs[FROM], Hostname);
	fprintf (tfile, "%s%s\n", Hdrs[TO], newto != NULLCHAR ? newto : to);
	fprintf (tfile, "%s%s\n\n", Hdrs[SUBJECT],
			 bounce ? "Failed mail" : "Return receipt");
	if (bounce)
		{
		fprintf (tfile, "  ===== transcript follows =====\n\n");
		for (; lp != NULLLIST; lp = lp->next)
			fprintf (tfile, "%s\n", lp->val);

		fputc ('\n', tfile);
		}

	fprintf (tfile, "  ===== %s follows ====\n",
			 bounce ? "Unsent message" : "Message header");

	last_had_newline = TRUE;
	while (fgets (buf, LINELEN, data) != NULLCHAR)
		{
		if (last_had_newline && buf[0] == '\n')
			break;
		fputs (buf, tfile);
		last_had_newline = (strchr (buf, '\n') != NULL);
		pwait (NULL);	/*  GeK - assume write slow, give up some time  */
		}

	if (bounce)
		{
		fputc ('\n', tfile);
		while ((cnt = fread (buf, 1, LINELEN, data)) > 0)
			{
			fwrite (buf, 1, cnt, tfile);
			pwait (NULL);	/*  GeK - assume write slow, give up some time  */
			}
		}

	fseek (tfile, 0L, 0);

	/* A null From<> so no looping replys to MAIL-DAEMONS */
#ifdef	notdef	/*  GeK - do use null path  */
	(void) sprintf (buf, "postmaster@%s", Hostname);
#endif
	if (mailuser (tfile, "", newto != NULLCHAR ? newto : to))
		file_error = TRUE;
	if (ferror (tfile))
		file_error = TRUE;
	if (fclose (tfile) == EOF)
		file_error = TRUE;
	free (newto);
	free (buf);
	return (file_error);
}

/* Check to see that we are accepting mail from this sender. If we aren't
 * then tell 'em that mail from them is Unwanted. I added this facility to
 * allow auto-rejection of messages from auto-reply daemons. G7LEU
 */
static int
validate_sender (char *sender, char **matched_rule)
{
	FILE *fp;
	char *buf;
	int	count = 0;
	int	kill = 0;

	if ((fp = fopen (Mailkill, READ_TEXT)) == NULLFILE)
		/* Can't open SMTP reject file so let 'em through */
		return 1;

	buf = mallocw (LINELEN);
	while (fgets (buf, LINELEN, fp) != NULLCHAR)
		{
		/*  GeK - assume wildmat() is slow so give up some time  */
		if (count++ == 10)
			{
			pwait (NULL);
			count = 0;
			}

		/*  Skip comments and blank lines  */
		if (buf[0] == '#' || buf[0] == '\n')
			continue;

		rip (buf);
		strlwr (buf);	/*  GeK - force lower case for pattern match  */
		if (buf[0] == '!')
			{
			/*  Keep match?  */
			if (wildmat (sender, buf+1, NULLCHARP))
				{
				free (*matched_rule);		/*  Reset any bad sender rule  */
				*matched_rule = NULL;
				fclose (fp);
				free (buf);
				return (1);		/*  Sender OK  */
				}
			}
		else
			{
			if (kill == 0 && wildmat (sender, buf, NULLCHARP))
				{
				/*  Matched bad sender, remember rule  */
				free (*matched_rule);
				*matched_rule = strdup (buf);
				kill = 1;
				/*  Can't return here until checked all ! rules  */
				}
			}

		}

	fclose (fp);
	free (buf);
	return (kill ? 0 : 1);
}

/* Check to see if a local user actually exists - use the FTPUSERS file
 * Returns 0 if the user doesn't exist, 1 otherwise
 */
static int
validate_user (char *user)
{
	char *cp;

/* always allow 'postmaster' */
	if (stricmp (user, "postmaster") == 0)
		{
		return 1;
		}
/* otherwise check the ftpusers file */

#	ifdef NOVELL

	if (novell_start)
		{
		if (!get_mail_id (user, novell_server_name))
			{
			return 0;
			}

		}
	else
#	endif
		{
		if ((cp = userlookup (user, NULL, NULLCHARP, NULLINT, NULL)) == NULLCHAR)
			{
			return 0;
			}

		free (cp);
		}

	return 1;
}

/*  GeK - added 'smtp server ...' commands  */
int
doserver (int argc, char *argv[], void *p)
{
	return subcmd (smtp_server_cmds, argc, argv, p);
}

#ifdef	COMMIT_CACHE
static int
docommit (int argc, char *argv[], void *p)
{
	return setbool (&commit_cache, "Cache commit", argc, argv);
}
#endif	/* COMMIT_CACHE */

static int
doreceipt (int argc, char *argv[], void *p)
{
	return setbool (&smtp_receipt, "Return receipt", argc, argv);
}

static int
dorelay (int argc, char *argv[], void *p)
{
	return setbool (&smtp_relay, "Relaying", argc, argv);
}

#ifdef	SERVER_TRACE
static int
do_server_trace (int argc, char *argv[], void *p)
{
	return setint (&smtp_trace, "SMTP server trace", argc, argv);
}
#endif	/* SERVER_TRACE */
